/* hostproc - SNMP agent for monitoring detailed list of running processes 
   Copyright (C) 2020 Sergey Poznyakoff

   Hostproc is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3 of the License, or (at your
   option) any later version.

   Hostproc is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with hostproc. If not, see <http://www.gnu.org/licenses/>. */
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>
#include "hostproc.h"

struct tokbuf {
	char const *val;
	char const *cur;
	char *tokptr;
	size_t toksize;
	int reuse;
};

static void
tokbuf_init(struct tokbuf *tb, char const *val)
{
	tb->val = tb->cur = val;
	tb->tokptr = NULL;
	tb->toksize = 0;
	tb->reuse = 0;
}

static void
tokbuf_free(struct tokbuf *tb)
{
	free(tb->tokptr);
}

static char *
tokbuf_morespace(struct tokbuf *tb, size_t len)
{
	while (len > tb->toksize) {
		char *p = a2nrealloc(tb->tokptr, &tb->toksize, 1);
		if (!p)
			abend_nomem();
		tb->tokptr = p;
	}
	return tb->tokptr;
}

static int
tokbuf_eof(struct tokbuf *tb)
{
	if (tb->reuse)
		return 0;
	tb->cur += strspn(tb->cur, " \t");
	return *tb->cur == 0;
}

static char *
getquote(struct tokbuf *tb)
{
	size_t len;
	size_t nesc = 0;
	char *ret, *p;

	tb->cur++; /* Skip initial quote character */
	/* Scan the line to determine its length */
	for (len = 0; tb->cur[len] != '"'; len++) {
		if (!tb->cur[len]) {
			config_perror("unterminated quoted string near");
			return NULL;
		}
		if (tb->cur[len] == '\\') {
			nesc++;
			len++;
			if (!tb->cur[len]) {
				config_perror("unterminated quoted string near");
				return NULL;
			}
		}
	}

	ret = tokbuf_morespace(tb, len - nesc + 1);
	p = ret;
	while (*tb->cur != '"') {
		if (*tb->cur == '\\')
			tb->cur++;
		*p++ = *tb->cur++;
	}
	*p = 0;
	tb->cur++;
	
	return ret;
}
	
static char *
getstr(struct tokbuf *tb)
{
	size_t len;
	char *ret;

	if (tb->reuse) {
		tb->reuse = 0;
		return tb->tokptr;
	}
	
	len = strspn(tb->cur, " \t");
	tb->cur += len;
	if (*tb->cur == 0) {
		config_perror("expected string, but found end of line");
		return NULL;
	}
	if (*tb->cur == '"') 
		return getquote(tb);
	len = strcspn(tb->cur, " \t");

	ret = tokbuf_morespace(tb, len + 1);
	memcpy(ret, tb->cur, len);
	ret[len] = 0;
	tb->cur += len;

	return ret;
}

static int
convnum(char *val, int issize, unsigned long *retval)
{
	unsigned long n;
	char *p;
	
	n = strtoul(val, &p, 10);
	if (n == UINT_MAX && errno == ERANGE) {
		config_perror("bad number");
		return -1;
	}
	if (*p) {
		unsigned long factor;
		if (!issize) {
			config_perror("bad number");
			return -1;
		}
		
		switch (*p) {
		case 0:
			factor = 1;
			break;
		case 'k':
		case 'K':
			factor = 1024UL;
			break;
		case 'm':
		case 'M':
			factor = 1024*1024UL;
			break;
		case 'g':
		case 'G':
			factor = 1024*1024*1024UL;
			break;
		default:
			config_perror("bad number");
			return -1;
		}

		if (ULONG_MAX / factor < n) {
			config_perror("bad number");
			return -1;
		}
		n *= factor;
	}
	
	*retval = n;
	return 0;
}

static int
getnum(struct tokbuf *tb, unsigned long *retval, char const *what)
{
	unsigned long n;
	char const *input;
	char *p;
	
	if (tb->reuse) {
		input = tb->tokptr;
	} else {
		tb->cur += strspn(tb->cur, " \t");
		if (*tb->cur == 0) {
			config_perror("expected number, but found end of line");
			return -1;
		}
		input = tb->cur;
	}
	n = strtoul(input, &p, 10);
	if ((n == UINT_MAX && errno == ERANGE) ||
	    (*p && !(*p == ' ' || *p == '\t'))) {
		char errbuf[128];
		snprintf(errbuf, sizeof(errbuf), "bad %s", what);
		config_perror(errbuf);
		tb->reuse = 0;
		return -1;
	}
	*retval = n;
	if (tb->reuse)
		tb->reuse = 0;
	else
		tb->cur = p;
	return 0;
}

static int
getnum_dfl(struct tokbuf *tb, unsigned long *retval, char const *what,
	   unsigned long dfl)
{
	if (tokbuf_eof(tb)) {
		*retval = dfl;
		return 0;
	}
	return getnum(tb, retval, what);
}

static inline void
putback(struct tokbuf *tb)
{
	tb->reuse = 1;
}

static char const *
basename(char const *str)
{
	char const *p = strrchr(str, '/');
	return p ? p + 1 : str;
}
	
static int
cmp_basename(PROCGROUP *pgr, char const *str)
{
	return strcmp(basename(pgr->pattern), basename(str));
}

static int
cmp_exact(PROCGROUP *pgr, char const *str)
{
	return strcmp(pgr->pattern, str);
}

static int
cmp_rx(PROCGROUP *pgr, char const *str)
{
	return regexec(&pgr->rx, str, 0, NULL, 0);
}

static int (*cmpfun[])(PROCGROUP *, char const *str) = {
	cmp_exact,
	cmp_basename,
	cmp_rx
};

int
procgroup_match(PROCESS *proc, PROCGROUP *pgr)
{
	char *str;

	switch (pgr->field) {
	case pf_comm:
		str = proc->comm;
		break;
	case pf_exe:
		str = proc->exe;
		break;
	case pf_cmdline:
		str = proc->cmdline;
		break;
	default:
		abort();
	}
	if (!str)
		return 1;
	return cmpfun[pgr->match](pgr, str);
}

PROCGROUP *
procgroup_find(PROCGROUP_LIST *pgrlist, char const *name)
{
	PROCGROUP *pgr;
	PROCGROUP_LIST_FOREACH(pgr, pgrlist) {
		if (strcmp(pgr->name, name) == 0)
			return pgr;
	}
	return NULL;
}

PROCGROUP *
procgroup_list_match_first(PROCESS *proc, PROCGROUP_LIST *pgrlist)
{
	PROCGROUP *ent;

	PROCGROUP_LIST_FOREACH(ent, pgrlist) {
		if (procgroup_match(proc, ent) == 0)
			return ent;
	}
	return NULL;
}

PROCGROUP *
procgroup_match_next(PROCESS *proc, PROCGROUP *ent)
{
	while (ent && (ent = PROCGROUP_NEXT(ent)) != NULL) {
		if (procgroup_match(proc, ent) == 0)
			break;
	}
	return ent;
}

PROCGROUP *
procgroup_create(char const *name)
{
	PROCGROUP *ent;

	if ((ent = calloc(1, sizeof(*ent))) == NULL
	    || (ent->name = strdup(name)) == NULL
	    || (ent->pattern = strdup(name)) == NULL)
		abend_nomem();
	
	return ent;
}

void
procgroup_free(PROCGROUP *ent)
{
	free(ent->name);
	free(ent->pattern);
	if (ent->match == match_rx)
		regfree(&ent->rx);
	free(ent);
}

static void
procgroup_compile_rx(PROCGROUP *pgr)
{
	int ec = regcomp(&pgr->rx, pgr->pattern, REG_EXTENDED|REG_NOSUB);
	if (ec) {
		char errbuf[512];
		regerror(ec, &pgr->rx, errbuf, sizeof(errbuf));
		config_perror(errbuf);
		pgr->match = match_exact;
	}
}

static void
procgroup_set_match(PROCGROUP *pgr, int match)
{
	if (pgr->match != match) {
		if (pgr->match == match_rx)
			regfree(&pgr->rx);
		pgr->match = match;
		if (pgr->match == match_rx)
			procgroup_compile_rx(pgr);
	}
}

static int skip_group; /* Skip group sub-statements until the next "group"
			  or "exclude" statement. */
static PROCGROUP_LIST *cur_group_list;

static PROCGROUP *
last_procgroup(void)
{
	if (skip_group)
		return NULL;
	if (!cur_group_list || PROCGROUP_LIST_EMPTY(cur_group_list)) {
		config_perror("no previous \"group\" or \"exclude\" statement");
		skip_group = 1;
		return NULL;
	}
	return PROCGROUP_LIST_LAST(cur_group_list);
}

static void
clear_last_procgroup(void)
{
	cur_group_list = NULL;
	skip_group = 0;
}

static void
parse_general_group(PROCGROUP_LIST *plist, const char *kw, char *val)
{
	PROCGROUP *pgr;

	pgr = procgroup_find(plist, val);
	if (pgr) {
		config_perror("group already defined");
		cur_group_list = NULL;
		skip_group = 1;
	} else {
		pgr = procgroup_create(val);
		if (!pgr)
			abend_nomem();
		PROCGROUP_LIST_APPEND(plist, pgr);
		cur_group_list = plist;
		skip_group = 0;
	}	
}

static void
parse_config_min(const char *kw, char *val)
{
	unsigned long n;
	PROCGROUP *grp = last_procgroup();
	if (grp && convnum(val, 0, &n) == 0) {
		grp->min_count = n;
	}
}

static void
parse_config_max(const char *kw, char *val)
{
	unsigned long n;
	PROCGROUP *grp = last_procgroup();
	if (grp && convnum(val, 0, &n) == 0) {
		grp->max_count = n;
	}
}

static void
parse_config_field(const char *kw, char *val)
{
	PROCGROUP *grp = last_procgroup();
	if (grp) {
		if (strcmp(val, "comm") == 0)
			grp->field = pf_comm;
		else if (strcmp(val, "exe") == 0)
			grp->field = pf_exe;
		else if (strcmp(val, "cmdline") == 0)
			grp->field = pf_cmdline;
		else
			config_perror("unrecognized field name; "
				      "expected one of: comm, exe, cmdline");
	}
}

static void
parse_config_pattern(const char *kw, char *val)
{
	PROCGROUP *grp = last_procgroup();
	if (grp) {
		free(grp->pattern);
		if (!(grp->pattern = strdup(val)))
			abend_nomem();
		if (grp->match == match_rx)
			procgroup_compile_rx(grp);
	}
}

static void
parse_config_match(const char *kw, char *val)
{
	PROCGROUP *grp = last_procgroup();
	if (grp) {
		if (strcmp(val, "exact") == 0)
			procgroup_set_match(grp, match_exact);
		else if (strcmp(val, "basename") == 0)
			procgroup_set_match(grp, match_basename);
		else if (strcmp(val, "rx") == 0)
			procgroup_set_match(grp, match_rx);
		else
			config_perror("unrecognized match type; "
				      "expected one of: exact, basename, rx");
	}
}

static void
parse_config_vsize(const char *kw, char *val)
{
	PROCGROUP *grp = last_procgroup();
	if (grp) {
		unsigned long n;
		if (convnum(val, 1, &n) == 0)
			grp->vsize_max = n;
	}
}

static void
parse_config_rss(const char *kw, char *val)
{
	PROCGROUP *grp = last_procgroup();
	if (grp) {
		unsigned long n;
		if (convnum(val, 1, &n) == 0)
			grp->rss_max = n;
	}
}

static void
parse_config_cpu(const char *kw, char *val)
{
	PROCGROUP *grp = last_procgroup();
	if (grp) {
		unsigned long n;
		if (convnum(val, 0, &n) == 0) {
			if (n < 100)
				grp->cpu_max = n;
			else
				config_pwarn("CPU percentage out of range");
		}
	}
}
		
static void
parse_config_group(const char *kw, char *val)
{
	parse_general_group(&process_groups, kw, val);
}


static void
parse_config_exclude(const char *kw, char *val)
{
	parse_general_group(&exclude_groups, kw, val);
}

static void
parse_config_proc(const char *kw, char *val)
{
	char *p;
	PROCGROUP *ent;
	struct tokbuf tb;

	clear_last_procgroup();
	
	tokbuf_init(&tb, val);

	p = getstr(&tb);
	if (!p)
		return;
	ent = procgroup_create(p);
	if (!(ent->pattern = strdup(ent->name)))
		abend_nomem();
	if (tokbuf_eof(&tb)) {
		/*
		 * If neither MAX nor MIN are specified, they will default
		 * to infinity and 1 respectively ("at least one").
		 */
		ent->min_count = 1;
		ent->max_count = 0;
	} else if (getnum_dfl(&tb, &ent->max_count, "max count", 0) == 0
		   && getnum_dfl(&tb, &ent->min_count, "min count", 0) == 0) {
		if (!tokbuf_eof(&tb))
			config_pwarn("garbage after maximum count ignored");
		PROCGROUP_LIST_APPEND(&process_groups, ent);
	} else
		procgroup_free(ent);
	
	tokbuf_free(&tb);
}

static void
parse_config_ps_interval(const char *kw, char *val)
{
	unsigned long n;
	struct tokbuf tb;
	
	clear_last_procgroup();

	tokbuf_init(&tb, val);

	if (getnum(&tb, &n, "interval"))
		return;
	if (n > UINT_MAX) {
		config_perror("interval out of allowed range");
		return;
	}
	ps_interval = n;
	tokbuf_free(&tb);
}

static void
parse_config_pidfile(const char *kw, char *val)
{
	pidfile = strdup(val);
}

struct hostproc_mib_config {
        char *token;
        char *help;
        void (*parser)(const char *token, char *line);
        void (*releaser)(void);
	void (*fini)(void);
};

static struct hostproc_mib_config config[] = {
	{ "psInterval", "NUMBER",
	  parse_config_ps_interval },
	{ "pidfile", "FILE",
	  parse_config_pidfile },
	{ "proc", "NAME [MAX] [MIN]",
	  parse_config_proc },
	{ "group", "NAME",
	  parse_config_group },
	{ "exclude", "NAME",
	  parse_config_exclude },
	{ "min", "NUM",
	  parse_config_min },
	{ "max", "NUM",
	  parse_config_max },
	{ "field", "comm|exe|cmdline",
	  parse_config_field },
	{ "pattern", "STRING",
	  parse_config_pattern },
	{ "match", "exact|basename|rx",
	  parse_config_match },
	{ "vsize", "SIZE",
	  parse_config_vsize },
	{ "rss", "SIZE",
	  parse_config_rss },
	{ "cpu", "PCT",
	  parse_config_cpu },
	{ NULL }
};
	 
void
hostproc_init_config(void)
{
	struct hostproc_mib_config *cp;
        for (cp = config; cp->token; cp++) {
                if (!register_app_config_handler(cp->token,
						 cp->parser,
						 cp->releaser,
						 cp->help))
                        snmp_log(LOG_ERR,"can't register %s config handler\n",
				 cp->token);
        }
}

void
hostproc_config_finalize(void)
{
	struct hostproc_mib_config *cp;
        for (cp = config; cp->token; cp++) {
		if (cp->fini)
			cp->fini();
	}
}

void
hostproc_mib_config_help(void)
{
        int i;

        for (i = 0; config[i].token; i++) {
                printf("%s %s\n", config[i].token, config[i].help);
        }
}

